#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <ctype.h>
#include <stdlib.h>
#include "depacks.h"

#ifdef __GNUC__
#define PACK( __Declaration__ ) __Declaration__ __attribute__((__packed__))
#endif

#ifdef _MSC_VER
#define PACK( __Declaration__ ) __pragma( pack(push, 1) ) __Declaration__ __pragma( pack(pop))
#endif

int error(int error, const char* format, ...)
{
  va_list args;

  va_start(args, format);
  vfprintf(stderr, format, args);
  return error;
}

int errorno(int error, int eno, const char* format, ...)
{
  va_list args;

  va_start(args, format);
  vfprintf(stderr, format, args);
  fprintf(stderr, " (%d): %s\n", eno, strerror(eno));
  return error;
}

int readX(FILE* file, long offset, size_t n, void* out)
{

  if (fseek(file, offset, SEEK_SET))
    return errno;

  if (fread(out, n, 1, file) != 1)
    return errno;

  return 0;
}

int read16(FILE* file, long offset, uint16_t* out)
{
  return readX(file, offset, 2, out);
}

int read32(FILE* file, long offset, uint32_t* out)
{
  return readX(file, offset, 4, out);
}

PACK(struct section_t
{
  char name[8];
  uint32_t vsize;
  uint32_t rva;
  uint32_t rsize;
  uint32_t offset;
  uint32_t ptr_reloc;
  uint32_t ptr_line_no;
  uint16_t no_reloc;
  uint16_t no_line_no;
  uint32_t characteristics;
});

int read_section(FILE* file, long offset, struct section_t* out)
{
  return readX(file, offset, sizeof(struct section_t), out);
}

#define MZ_SIGNATURE_OFFSET 0
#define MZ_SIGNATURE 0x5a4d
#define PE_HEADER_OFFSET 0x3c

#define PE_SIGNATURE_OFFSET(x) ( x + 0)
#define PE_SIGNATURE 0x4550
#define OPTIONAL_HEADER_SIZE_OFFSET(x) ( x + 0x14)
#define OPTIONAL_HEADER_OFFSET(x) ( x + 0x18)
#define SECTIONS_NO_OFFSET(x) ( x + 0x6)

int get_sections_table(FILE* file, uint32_t* offset, uint16_t* sections_no)
{
    uint16_t mz_sign, opt_header_size;
    uint32_t pe_offset, pe_sign;

    /* MZ header */

    if (read16(file, MZ_SIGNATURE_OFFSET, &mz_sign))
      return errorno(1, errno, "Cannot read MZ signature");
    if (mz_sign != MZ_SIGNATURE)
      return error(2, "MZ signature is invalid. Found %04x expected %04x\n", mz_sign, MZ_SIGNATURE);
    if (read32(file, PE_HEADER_OFFSET, &pe_offset))
      return errorno(3, errno, "Cannot read PE header offset");

    /* PE header */
    if (read32(file, PE_SIGNATURE_OFFSET(pe_offset), &pe_sign))
      return errorno(4, errno, "Cannot read PE signature");
    if (pe_sign != PE_SIGNATURE)
      return error(5, "PE signature is invalid. Found %08x expected %08x\n", pe_sign, PE_SIGNATURE);
    if (read16(file, OPTIONAL_HEADER_SIZE_OFFSET(pe_offset), &opt_header_size))
      return errorno(6, errno, "Cannot read the size of the optional header");

    *offset = OPTIONAL_HEADER_OFFSET(pe_offset) + opt_header_size;

    if (read16(file, SECTIONS_NO_OFFSET(pe_offset), sections_no))
      return errorno(7, errno, "Cannot read the number of sections in the section table");

    return 0;
}



PACK(struct chunk_t 
{
  uint16_t signature; /* The word JJ (0x4a4a) */
  uint8_t padding;    /* Empty space to the next item (in DWORD) */
  uint8_t flags;      /* Flags*/
  uint32_t xorKey;    /* XORed with the first DWORD of the decoded data */
  uint32_t id;        /* An unique ID */
  uint32_t offset;    /* RVA of the encoded data */
  uint32_t size;      /* Size of the encoded data */
});

#define CHUNK_SIGNATURE 0x4a4a
#define CHUNK_FLAG_IGNORE 0x2
#define CHUNK_FLAG_COMPRESSED 0x1

int rva_to_offset(struct section_t* secs, uint16_t secno, uint32_t rva, uint32_t* offset)
{
  int i;
  for (i = 0; i < secno; i++)
  {
      if (rva >= secs[i].rva && rva < secs[i].rva + secs[i].vsize)
      {
          *offset = rva - secs[i].rva + secs[i].offset;
          return 0;
      }
  }

  return 1;
}

int main(int argc, char** argv)
{
  FILE* file = NULL;
  int retval = 0;

  uint32_t sec_table;
  uint16_t sec_no;

  struct section_t* sections = NULL;
  struct chunk_t chunk;
  uint32_t chunk_offset;

  char* buffer = NULL, *dest = NULL;

  char fileName[9];
  FILE* fout;
  uint32_t s = 0;
  uint32_t* ptr, *dptr;
  uint32_t len;

  if (argc != 2)
    return error(1, "Usage: %s PE_FILE\n", argv[0]);

  if (!(file = fopen(argv[1], "r+b")))  
    return errorno(2, errno, "Cannot open PE file %s", argv[1]);

  if ((retval = get_sections_table(file, &sec_table, &sec_no)))
  {
    retval = 30 + retval;
    goto end;
  }

  sections = malloc(sizeof(struct section_t)*sec_no);
  if ( ! sections)
  {
      retval = errorno(4, errno, "Cannot allocate memory for the sections table");
      goto end;  
  }
  
  if (readX(file, sec_table, sizeof(struct section_t)*sec_no, sections))
  {
      retval = errorno(5, errno, "Cannot read the sections table");
      goto end;  
  }  

  sec_table += 0x28 + sec_no*0x28;

  /* Process JJs */
  while (1)
  {
    if (readX(file, sec_table, sizeof(struct chunk_t), &chunk))
    {
      retval = errorno(6, errno, "Cannot read JJ chunk");
      goto end;        
    }

    if (chunk.signature != CHUNK_SIGNATURE)
      break;

    printf("-- JJ Chunk --\nId: %08x\nFlags: %s %s\nRVA: %08x\nSize: %x\nXOR key: %08x\nPadding: %u\n\n", 
          chunk.id, 
          chunk.flags & CHUNK_FLAG_IGNORE ? "IG" : "",
          chunk.flags & CHUNK_FLAG_COMPRESSED ? "CM": "",
          chunk.offset, chunk.size, chunk.xorKey, chunk.padding);

    if (chunk.flags & CHUNK_FLAG_IGNORE)
      goto next_chunk;

    if ((retval = rva_to_offset(sections, sec_no, chunk.offset, &chunk_offset)))
    {
      retval = 70 + retval;
      goto end;
    }

    buffer = malloc(chunk.size);
    if ( ! buffer)
    {
      retval = errorno(8, errno, "Cannot allocate a buffer for the chunk (%u)", chunk.size);
      goto end;
    }

    dest = malloc(chunk.size+1);
    if ( ! dest)
    {
      retval = errorno(8, errno, "Cannot allocate a buffer for the decoded chunk (%u)", chunk.size+1);
      goto end;
    }

    if (readX(file, chunk_offset, chunk.size, buffer))
    {
      retval = errorno(8, errno, "Cannot read the chunk");
      goto end;
    }

    if (chunk.flags & CHUNK_FLAG_COMPRESSED)
    {
      aP_depack_safe(buffer, chunk.size, dest, chunk.size);
    }
    else
    {
        s = chunk.xorKey;
        ptr = (uint32_t*)buffer;
        dptr = (uint32_t*)dest;
        len = chunk.size / 4;
		
        while (len--)
        {
            *dptr = *ptr - s;
            s += *dptr;
            dptr++;
            ptr++;
        }

        if (chunk.size & 0x3)
          memcpy(dptr, ptr, chunk.size & 0x3);
    }

    dest[chunk.size] = 0;
    if (chunk.size >= 4 && (chunk.flags & CHUNK_FLAG_COMPRESSED))
      *((uint32_t*)dest) ^= chunk.xorKey;

    sprintf(fileName, "%08x", chunk.id);

    fout = fopen(fileName, "wb");
    fwrite(dest, chunk.size+1, 1, fout);
    fclose(fout);

    free(buffer);
    buffer = NULL;
    free(dest);
    dest = NULL;
    
next_chunk:
    sec_table += 0x14 + chunk.padding*4;
    
  }

end:
  if (file)
    fclose(file);
  if (sections)
    free(sections);
  if (buffer)
    free(buffer);
  if (dest)
    free(dest);

  return retval;
}
